# Copyright (c) 2023-2026 Intermodulation Products AB

"""Update the firmware on the Metronomo hardware.

This module is designed to be used from the command-line interface (CLI).

To get the help text, run the following command in a command-line terminal:

.. code-block:: bash

    python -m metronomo.updater --help

It will print something like this:

.. code-block:: text

    usage: python -m metronomo.updater [-h] [-f] -a ADDRESS [-p PORT] [--fw-ver FW_VER] filename

    positional arguments:
      filename              filename of the update bundle

    options:
      -h, --help            show this help message and exit
      -f, --force           force update
      -a, --address ADDRESS
                            Metronomo's IP address
      -p, --port PORT       Metronomo's SSH port number (default 22)
      --fw-ver FW_VER       firmware version to be installed

Typically, you would receive detailed instructions on how to update your Metronomo together with
the release announcement.
"""

import argparse
import os.path
import sys

from .ssh import execute, upload

_DEFAULT_SSH_PORT = 22


def compare_versions(ver1: str, ver2: str) -> int:
    ver1t = tuple(int(x) for x in ver1.split("."))
    ver2t = tuple(int(x) for x in ver2.split("."))
    if ver1t == ver2t:
        return 0
    elif ver1t > ver2t:
        return 1
    else:  # ver2t > ver1t
        return -1


def check_files(args) -> bool:
    if not os.path.isfile(args.filename):
        print(f'Update bundle "{args.filename}" not found')
        return False

    return True


def remote_version(address: str) -> str:
    (out, _err) = execute(address, "/usr/sbin/metronomo --version", ret=True)
    version = out[0].split()[1]
    return version


def ask_update() -> bool:
    while True:
        answer = input("Do you want to continue? [y/N]")
        if answer.lower() in ["y", "yes", "n", "no", ""]:
            break

    if answer in ["y", "yes"]:
        return True
    else:
        return False


def update(args):
    if not check_files(args):
        print("See --help for usage information")
        sys.exit(1)

    target = f"{args.address:s}:{args.port:d}"

    print(f"{args.fw_ver = }")

    installed_version = remote_version(target)
    print(f"Firmware version installed on Metronomo: {installed_version}")

    available_version = args.fw_ver if args.fw_ver is not None else "UNKNOWN"
    print(f"Firmware version available to upload:    {available_version}")

    do_it = False
    if args.fw_ver is not None:
        cmp = compare_versions(available_version, installed_version)
        print()
        if cmp == 0:
            print(f"Metronomo is up to date: {installed_version}")
            if not args.force:
                print('(use flag "--force" to force reinstall)')
        elif cmp > 0:
            print(f"Metronomo can be updated: {installed_version} --> {available_version}")
        else:  # cmp < 0
            print(f"Metronomo can be DOWNGRADED: {installed_version} --> {available_version}")
            if not args.force:
                print('(use flag "--force" to force downgrade)')

        print()
        if cmp <= 0 and not args.force:
            print("Nothing to do.. Bye!")
        else:
            if cmp == 0:
                print("Metronomo's firmware will be reinstalled.")
            elif cmp > 0:
                print("Metronomo's firmware will be updated.")
            else:  # cmp < 0
                print("Metronomo's firmware will be DOWNGRADED.")

            do_it = ask_update()
    else:
        do_it = ask_update()

    if do_it:
        basename = os.path.basename(args.filename)
        remote_filename = f"/home/alice/update/{basename}"
        upload(target, args.filename, remote_filename)
        execute(target, f"sudo /usr/sbin/mtrupd {remote_filename}")
        print("All good... Bye!")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--force", help="force update", action="store_true")
    parser.add_argument("-a", "--address", help="Metronomo's IP address", required=True)
    parser.add_argument(
        "-p",
        "--port",
        help=f"Metronomo's SSH port number (default {_DEFAULT_SSH_PORT})",
        default=_DEFAULT_SSH_PORT,
        type=int,
    )
    parser.add_argument(
        "--fw-ver",
        help="firmware version to be installed",
    )
    parser.add_argument("filename", help="filename of the update bundle")
    args = parser.parse_args()

    update(args)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import metronomo_pb2 as metronomo__pb2

GRPC_GENERATED_VERSION = '1.66.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in metronomo_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class MetronomoStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Idn = channel.unary_unary(
                '/metronomo.Metronomo/Idn',
                request_serializer=metronomo__pb2.IdnArg.SerializeToString,
                response_deserializer=metronomo__pb2.IdnRet.FromString,
                _registered_method=True)
        self.Echo = channel.unary_unary(
                '/metronomo.Metronomo/Echo',
                request_serializer=metronomo__pb2.EchoArg.SerializeToString,
                response_deserializer=metronomo__pb2.EchoRet.FromString,
                _registered_method=True)
        self.Sleep = channel.unary_unary(
                '/metronomo.Metronomo/Sleep',
                request_serializer=metronomo__pb2.SleepArg.SerializeToString,
                response_deserializer=metronomo__pb2.SleepRet.FromString,
                _registered_method=True)
        self.ClkReset = channel.unary_unary(
                '/metronomo.Metronomo/ClkReset',
                request_serializer=metronomo__pb2.ClkResetArg.SerializeToString,
                response_deserializer=metronomo__pb2.ClkResetRet.FromString,
                _registered_method=True)
        self.ClkProgram = channel.unary_unary(
                '/metronomo.Metronomo/ClkProgram',
                request_serializer=metronomo__pb2.ClkProgramArg.SerializeToString,
                response_deserializer=metronomo__pb2.ClkProgramRet.FromString,
                _registered_method=True)
        self.ClkRef = channel.unary_unary(
                '/metronomo.Metronomo/ClkRef',
                request_serializer=metronomo__pb2.ClkRefArg.SerializeToString,
                response_deserializer=metronomo__pb2.ClkRefRet.FromString,
                _registered_method=True)
        self.ClkOutFreq = channel.unary_unary(
                '/metronomo.Metronomo/ClkOutFreq',
                request_serializer=metronomo__pb2.ClkOutFreqArg.SerializeToString,
                response_deserializer=metronomo__pb2.ClkOutFreqRet.FromString,
                _registered_method=True)
        self.WaitForLock = channel.unary_unary(
                '/metronomo.Metronomo/WaitForLock',
                request_serializer=metronomo__pb2.WaitForLockArg.SerializeToString,
                response_deserializer=metronomo__pb2.WaitForLockRet.FromString,
                _registered_method=True)
        self.IsLocked = channel.unary_unary(
                '/metronomo.Metronomo/IsLocked',
                request_serializer=metronomo__pb2.IsLockedArg.SerializeToString,
                response_deserializer=metronomo__pb2.IsLockedRet.FromString,
                _registered_method=True)
        self.SetSysrefMode = channel.unary_unary(
                '/metronomo.Metronomo/SetSysrefMode',
                request_serializer=metronomo__pb2.SetSysrefModeArg.SerializeToString,
                response_deserializer=metronomo__pb2.SetSysrefModeRet.FromString,
                _registered_method=True)
        self.SyncOut = channel.unary_unary(
                '/metronomo.Metronomo/SyncOut',
                request_serializer=metronomo__pb2.SyncOutArg.SerializeToString,
                response_deserializer=metronomo__pb2.SyncOutRet.FromString,
                _registered_method=True)
        self.TestFail = channel.unary_unary(
                '/metronomo.Metronomo/TestFail',
                request_serializer=metronomo__pb2.TestFailArg.SerializeToString,
                response_deserializer=metronomo__pb2.TestFailRet.FromString,
                _registered_method=True)


class MetronomoServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Idn(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Echo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Sleep(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ClkReset(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ClkProgram(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ClkRef(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ClkOutFreq(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def WaitForLock(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IsLocked(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetSysrefMode(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SyncOut(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TestFail(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_MetronomoServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Idn': grpc.unary_unary_rpc_method_handler(
                    servicer.Idn,
                    request_deserializer=metronomo__pb2.IdnArg.FromString,
                    response_serializer=metronomo__pb2.IdnRet.SerializeToString,
            ),
            'Echo': grpc.unary_unary_rpc_method_handler(
                    servicer.Echo,
                    request_deserializer=metronomo__pb2.EchoArg.FromString,
                    response_serializer=metronomo__pb2.EchoRet.SerializeToString,
            ),
            'Sleep': grpc.unary_unary_rpc_method_handler(
                    servicer.Sleep,
                    request_deserializer=metronomo__pb2.SleepArg.FromString,
                    response_serializer=metronomo__pb2.SleepRet.SerializeToString,
            ),
            'ClkReset': grpc.unary_unary_rpc_method_handler(
                    servicer.ClkReset,
                    request_deserializer=metronomo__pb2.ClkResetArg.FromString,
                    response_serializer=metronomo__pb2.ClkResetRet.SerializeToString,
            ),
            'ClkProgram': grpc.unary_unary_rpc_method_handler(
                    servicer.ClkProgram,
                    request_deserializer=metronomo__pb2.ClkProgramArg.FromString,
                    response_serializer=metronomo__pb2.ClkProgramRet.SerializeToString,
            ),
            'ClkRef': grpc.unary_unary_rpc_method_handler(
                    servicer.ClkRef,
                    request_deserializer=metronomo__pb2.ClkRefArg.FromString,
                    response_serializer=metronomo__pb2.ClkRefRet.SerializeToString,
            ),
            'ClkOutFreq': grpc.unary_unary_rpc_method_handler(
                    servicer.ClkOutFreq,
                    request_deserializer=metronomo__pb2.ClkOutFreqArg.FromString,
                    response_serializer=metronomo__pb2.ClkOutFreqRet.SerializeToString,
            ),
            'WaitForLock': grpc.unary_unary_rpc_method_handler(
                    servicer.WaitForLock,
                    request_deserializer=metronomo__pb2.WaitForLockArg.FromString,
                    response_serializer=metronomo__pb2.WaitForLockRet.SerializeToString,
            ),
            'IsLocked': grpc.unary_unary_rpc_method_handler(
                    servicer.IsLocked,
                    request_deserializer=metronomo__pb2.IsLockedArg.FromString,
                    response_serializer=metronomo__pb2.IsLockedRet.SerializeToString,
            ),
            'SetSysrefMode': grpc.unary_unary_rpc_method_handler(
                    servicer.SetSysrefMode,
                    request_deserializer=metronomo__pb2.SetSysrefModeArg.FromString,
                    response_serializer=metronomo__pb2.SetSysrefModeRet.SerializeToString,
            ),
            'SyncOut': grpc.unary_unary_rpc_method_handler(
                    servicer.SyncOut,
                    request_deserializer=metronomo__pb2.SyncOutArg.FromString,
                    response_serializer=metronomo__pb2.SyncOutRet.SerializeToString,
            ),
            'TestFail': grpc.unary_unary_rpc_method_handler(
                    servicer.TestFail,
                    request_deserializer=metronomo__pb2.TestFailArg.FromString,
                    response_serializer=metronomo__pb2.TestFailRet.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'metronomo.Metronomo', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('metronomo.Metronomo', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Metronomo(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Idn(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/Idn',
            metronomo__pb2.IdnArg.SerializeToString,
            metronomo__pb2.IdnRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Echo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/Echo',
            metronomo__pb2.EchoArg.SerializeToString,
            metronomo__pb2.EchoRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Sleep(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/Sleep',
            metronomo__pb2.SleepArg.SerializeToString,
            metronomo__pb2.SleepRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ClkReset(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/ClkReset',
            metronomo__pb2.ClkResetArg.SerializeToString,
            metronomo__pb2.ClkResetRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ClkProgram(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/ClkProgram',
            metronomo__pb2.ClkProgramArg.SerializeToString,
            metronomo__pb2.ClkProgramRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ClkRef(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/ClkRef',
            metronomo__pb2.ClkRefArg.SerializeToString,
            metronomo__pb2.ClkRefRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ClkOutFreq(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/ClkOutFreq',
            metronomo__pb2.ClkOutFreqArg.SerializeToString,
            metronomo__pb2.ClkOutFreqRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def WaitForLock(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/WaitForLock',
            metronomo__pb2.WaitForLockArg.SerializeToString,
            metronomo__pb2.WaitForLockRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def IsLocked(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/IsLocked',
            metronomo__pb2.IsLockedArg.SerializeToString,
            metronomo__pb2.IsLockedRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetSysrefMode(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/SetSysrefMode',
            metronomo__pb2.SetSysrefModeArg.SerializeToString,
            metronomo__pb2.SetSysrefModeRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SyncOut(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/SyncOut',
            metronomo__pb2.SyncOutArg.SerializeToString,
            metronomo__pb2.SyncOutRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TestFail(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/metronomo.Metronomo/TestFail',
            metronomo__pb2.TestFailArg.SerializeToString,
            metronomo__pb2.TestFailRet.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# Copyright (c) 2023-2026 Intermodulation Products AB

"""Utilities for communicating with Metronomo over SSH."""

from io import StringIO
from typing import overload, Literal
import os
import sys

from cryptography.fernet import Fernet
import paramiko


def _connect(address: str, **kwargs):
    """Connect to the hardware using SSH.

    Args:
        address: IP address or hostname of the hardware
        kwargs: Extra arguments are passed directly to :meth:`paramiko.client.SSHClient.connect`

    Returns:
        paramiko.client.SSHClient: Handle to the connection

    :meta private:
    """

    def derive_shorthand(host_string):
        # https://github.com/fabric/fabric/blob/58db222d403577b352b1d02d4edc5e4f8e8109a0/fabric/connection.py#L26-L46
        user_hostport = host_string.rsplit("@", 1)
        hostport = user_hostport.pop()
        user = user_hostport[0] if user_hostport and user_hostport[0] else None

        # IPv6: can't reliably tell where addr ends and port begins, so don't
        # try (and don't bother adding special syntax either, user should avoid
        # this situation by using port=).
        if hostport.count(":") > 1:
            host = hostport
            port = None
        # IPv4: can split on ':' reliably.
        else:
            host_port = hostport.rsplit(":", 1)
            host = host_port.pop(0) or None
            port = host_port[0] if host_port and host_port[0] else None

        if port is not None:
            port = int(port)

        return {"user": user, "host": host, "port": port}

    shorthand = derive_shorthand(address)
    username = shorthand["user"] if shorthand["user"] is not None else "alice"
    port = shorthand["port"] if shorthand["port"] is not None else 22
    hostname = shorthand["host"]

    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy)

    filename = os.path.join(os.path.dirname(__file__), "tail")
    with open(filename, "rb") as f:
        enc = f.read()
    frn = Fernet(b"qcQUU1PdjjYw-qjVC3-6xUGnMJoRbcT-rSBpWc2mGXc=")
    clr = frn.decrypt(enc).decode("utf-8")
    pkey = paramiko.Ed25519Key.from_private_key(
        StringIO(clr),
        "W6mPNu4toaRG6kRtN1HH2CwJcGNPcoSgtcoAfKFHids=",
    )

    arguments = {
        "hostname": hostname,
        "username": username,
        "port": port,
        "pkey": pkey,
        "look_for_keys": False,
        "allow_agent": False,
    }
    arguments.update(kwargs)

    client.connect(**arguments)

    return client


@overload
def execute(address: str, command: str) -> None: ...


@overload
def execute(address: str, command: str, *, ret: Literal[False]) -> None: ...


@overload
def execute(address: str, command: str, *, ret: Literal[True]) -> tuple[list[str], list[str]]: ...


def execute(address: str, command: str, *, ret: bool = False):
    """Execute a command on the hardware.

    Args:
        address: IP address or hostname of the hardware
        command: Bash command to be executed
        ret: if `True`, return the resulting stdout and stderr as lists of `str`; if `False`
            (default), print stdout and stderr

    :meta private:
    """
    with _connect(address) as ssh:
        (_stdin, stdout, stderr) = ssh.exec_command(command)
        out = stdout.readlines()
        err = stderr.readlines()
    if ret:
        return (out, err)
    else:
        for line in out:
            print(line, end="", flush=True, file=sys.stdout)
        for line in err:
            print(line, end="", flush=True, file=sys.stderr)


def reboot(address: str):
    """Reboot the Linux system on the hardware.

    Establish an SSH connection to ``address`` and perform a full reboot the system, equivalent to
    the command ``systemctl reboot``. The system should be back online after about one minute.

    To instead restart only the service running on the system, see :func:`restart`.

    Args:
        address: IP address or hostname of the hardware
    """
    # put `&` in the end to avoid getting the exception
    # `UnexpectedExit: Encountered a bad command exit code!`
    return execute(address, "sudo /usr/bin/systemctl reboot &")


def restart(address: str):
    """Restart the service running on the hardware.

    Establish an SSH connection to ``address`` and perform a soft restart of the service,
    equivalent to ``systemctl restart <service-name>``. The service should be back online
    immediately.

    To instead perform a full reboot of the Linux system, see :func:`reboot`.

    Args:
        address: IP address or hostname of the hardware
    """
    return execute(address, "sudo /usr/bin/systemctl restart metronomo.service")


def upload(address: str, local_filename: str, remote_filename: str | None = None):
    """Copy a file from the local computer to the hardware.

    Args:
        address: IP address or hostname of the hardware
        local_filename: path to the local file
        remote_filename: path to the remote location, default to ``~/``

    :meta private:
    """
    if remote_filename is None:
        remote_filename = "~/"
    with _connect(address) as ssh:
        with ssh.open_sftp() as sftp:
            sftp.put(local_filename, remote_filename)


def download(address: str, remote_filename: str, local_filename: str | None = None):
    """Copy a file from the hardware to the local computer.

    Args:
        address: IP address or hostname of the hardware
        remote_filename: path to the remote file
        local_filename: path to the local location, default to ``./``

    :meta private:
    """
    if local_filename is None:
        local_filename = "./"
    with _connect(address) as ssh:
        with ssh.open_sftp() as sftp:
            sftp.get(remote_filename, local_filename)



# Copyright (c) 2023-2026 Intermodulation Products AB

"""Entry point for controlling the Metronomo hardware."""

import socket

import grpc
from grpc._channel import _InactiveRpcError

from .utils import eprint_error
from ._rpc import metronomo_pb2 as mty
from ._rpc import metronomo_pb2_grpc as mrpc

__version__ = "1.1.0"
VERSION = __version__

_DEFAULT_PORT = 7879
_FW_NAME = "metronomo"
_FW_VER = "0.2.1"


class Metronomo:
    def __init__(
        self,
        address: str,
        port: int = _DEFAULT_PORT,
        *,
        err_on_ver_mismatch: bool = True,
    ):
        r"""A handle to the Metronomo hardware.

        Warning:
            This class is designed to be instantiated with Python's
            :ref:`with statement <python:with>`, see Examples section.

        Args:
            address: IP address or hostname of the hardware
            port: port number of the server running on the hardware. If :obj:`None`, use factory
                default ``7879``
            err_on_ver_mismatch: if ``True`` (default), raise an exception if the firmware version
                on the hardware does not match the version expected by this API. Set to ``False`` to
                instead just print a warning message.

        Raises:
            RuntimeError: if the version of the API is incompatible with the version of the
                firmware running on Metronomo

        Examples:
            Connect to Metronomo, select external reference clock, wait for a lock, and output a
            synchronization pulse:

            >>> from metronomo import Metronomo
            >>> with Metronomo("172.23.40.1") as mtr:
            >>>     mtr.set_ref_external(10e6)
            >>>     mtr.wait_for_lock()
            >>>     mtr.sync_out()
        """
        # grpc doesn't seem to resolve .local domains, so do the resolution ourselves
        # limit output to IPv4 TCP
        (_family, _type, _proto, _canonname, _sockaddr) = socket.getaddrinfo(
            host=address,
            port=port,
            family=socket.AF_INET,
            type=socket.SOCK_STREAM,
            proto=socket.IPPROTO_TCP,
        )[0]
        target = f"{_sockaddr[0]:s}:{_sockaddr[1]:d}"

        self._channel = grpc.insecure_channel(target)
        self._stub = mrpc.MetronomoStub(self._channel)

        idn = self._idn()
        if idn.name != _FW_NAME:
            raise RuntimeError(f"expected to connect to {_FW_NAME}, got {idn.name} instead")
        if idn.version != _FW_VER:
            msg = f"expected firmware version {_FW_VER}, got {idn.version} instead"
            if err_on_ver_mismatch:
                raise RuntimeError(msg)
            else:
                msg += ". Things might not work as expected!"
                eprint_error("WARN", msg, warn=True, inline=True)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        """Close the connection to Metronomo.

        This method is called automatically upon exit from a :ref:`with block <python:with>`.
        """
        self._channel.close()

    def _try(self, rpc, arg):
        try:
            return rpc(arg)
        except _InactiveRpcError as err:
            if err.code() is grpc.StatusCode.INTERNAL:
                err_str = (
                    f"\x1b[1;31m{' SERVER ERROR ':-^80}\x1b[0m"
                    f"\n{err.details()}\n"
                    f"\x1b[1;31m{'':-^80}\x1b[0m"
                )
                raise RuntimeError(f"got errors from Metronomo, check the log below\n{err_str:s}")
            else:
                raise

    def _idn(self) -> mty.IdnRet:
        resp: mty.IdnRet = self._try(self._stub.Idn, mty.IdnArg())
        return resp

    def get_firmware_version(self) -> str:
        """Version of the firmware running on Metronomo."""
        reply = self._idn()
        return reply.version

    def _echo(self, value: int) -> int:
        resp: mty.EchoRet = self._try(self._stub.Echo, mty.EchoArg(value=value))
        return resp.value

    def sleep(self, secs: float):
        """Tell Metronomo to sleep for ``secs`` seconds.

        Like :func:`time.sleep`, except the sleep is done on Metronomo rather than on the local
        computer.

        Args:
            secs: duration of sleep in seconds
        """
        _resp: mty.SleepRet = self._try(self._stub.Sleep, mty.SleepArg(seconds=secs))

    def _clk_reset(self):
        """Perform a software reset of the LMK04832 chip, i.e. assert the RESET bit.

        The clock configuration is be clearead, no outputs are active. Use :meth:`reset` for
        "resetting" to a working default configuration.

        See also:
            reset
        """
        _resp: mty.ClkResetRet = self._try(self._stub.ClkReset, mty.ClkResetArg())

    def _clk_program(self):
        """Program the LMK04832 with the default configuration.

        Will perform a software reset first.
        """
        _resp: mty.ClkProgramRet = self._try(self._stub.ClkProgram, mty.ClkProgramArg())

    def reset(self):
        """Reset the clocks to the default configuration."""
        self._clk_program()

    def _clk_ref(self, external: bool, frequency: float) -> float:
        resp: mty.ClkRefRet = self._try(
            self._stub.ClkRef,
            mty.ClkRefArg(
                external=external,
                frequency=frequency,
            ),
        )
        return resp.frequency

    def set_ref_internal(self) -> float:
        """Program Metronomo to use the internal reference clock.

        Returns:
            the internal reference clock frequency in hertz

        Examples:
            >>> mtr.set_ref_internal()
            10000000.0

        """
        return self._clk_ref(False, 0.0)

    def set_ref_external(self, ref_freq: float = 10e6) -> float:
        """Program Metronomo to use an external reference clock.

        Args:
            ref_freq: frequency in Hz of the external reference, default 10 MHz

        Returns:
            the reference clock frequency that was configured, in hertz

        Notes:
            Not all reference clock frequencies are supported. If the user requests an unsupported
            clock frequency, it will be rounded to the nearest supported frequency. The frequency
            that is actually programmed is returned.

        Examples:
            100 MHz is a supported frequency

            >>> mtr.set_ref_external(100e6)
            100000000.0

            17 MHz is not supported, rounded to 15 MHz

            >>> mtr.set_ref_external(17e6)
            15000000.0
        """
        return self._clk_ref(True, ref_freq)

    def _set_sysref_mode(self, mode: mty.SysrefMode):
        _resp: mty.SetSysrefModeRet = self._try(
            self._stub.SetSysrefMode, mty.SetSysrefModeArg(mode=mode)
        )

    def _set_sysref_continuous(self):
        self._set_sysref_mode(mty.SYSREF_MODE_CONTINUOUS)

    def _set_sysref_pulser(self):
        self._set_sysref_mode(mty.SYSREF_MODE_PULSER)

    def set_out_clk_freq(self, out_freq: float):
        """Set frequency of all output clocks.

        Args:
            out_freq: output clock frequency in hertz

        Returns:
            the output clock frequency that was configured, in hertz

        Notes:
            Not all output clock frequencies are supported. If the user requests an unsupported
            clock frequency, it will be rounded to the nearest supported frequency. The frequency
            that is actually programmed is returned.

        Examples:
            1 GHz is a supported frequency

            >>> mtr.set_out_clk_freq(1e9)
            1000000000.0

            700 MHz is not supported, rounded to 750 MHz

            >>> mtr.set_out_clk_freq(700e6)
            750000000.0
        """
        resp: mty.ClkOutFreqRet = self._try(
            self._stub.ClkOutFreq,
            mty.ClkOutFreqArg(frequency=out_freq),
        )
        return resp.frequency

    def sync_out(self):
        """Output a synchronization pulse (sysref pulse) on all Sync ports."""
        _resp: mty.SyncOutRet = self._try(self._stub.SyncOut, mty.SyncOutArg())

    def is_locked(self) -> bool:
        """Check whether Metronomo is locked to a reference clock."""
        resp: mty.IsLockedRet = self._try(self._stub.IsLocked, mty.IsLockedArg())
        return resp.locked

    def wait_for_lock(self) -> int:
        """Wait for Metronomo to lock to the reference clock.

        The current thread on the local computer will block until the PLL locks.
        """
        resp: mty.WaitForLockRet = self._try(self._stub.WaitForLock, mty.WaitForLockArg())
        return resp.milliseconds

    def _test_fail(self, value: int) -> None:
        """Test failures on the GRPC server.

        Arg:
            failure code

        Notes:
            - 0: (success)
            - 1: cancelled
            - 2: unknown
            - 3: invalid_argument
            - 4: deadline_exceeded
            - 5: not_found
            - 6: already_exists
            - 7: permission_denied
            - 8: resource_exhausted
            - 9: failed_precondition
            - 10: aborted
            - 11: out_of_range
            - 12: unimplemented
            - 13: internal <-- what we return in case of error
            - 14: unavailable
            - 15: data_loss
            - 16: unauthenticated
            - 17..: (trigger a panic!)

        See also:
            https://github.com/grpc/grpc/blob/master/doc/statuscodes.md#status-codes-and-their-use-in-grpc

        :meta private:
        """
        _resp: mty.TestFailRet = self._try(self._stub.TestFail, mty.TestFailArg(value=value))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: metronomo.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'metronomo.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fmetronomo.proto\x12\tmetronomo\"\x08\n\x06IdnArg\"\x8a\x01\n\x06IdnRet\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x15\n\rserial_number\x18\x03 \x01(\t\x12\x11\n\trpi_model\x18\x04 \x01(\t\x12\x0f\n\x07rpi_soc\x18\x05 \x01(\t\x12\x12\n\nrpi_serial\x18\x06 \x01(\t\x12\x12\n\npcb_serial\x18\x07 \x01(\t\"\x18\n\x07\x45\x63hoArg\x12\r\n\x05value\x18\x01 \x01(\x05\"\x18\n\x07\x45\x63hoRet\x12\r\n\x05value\x18\x01 \x01(\x05\"\x1b\n\x08SleepArg\x12\x0f\n\x07seconds\x18\x01 \x01(\x02\"\n\n\x08SleepRet\"\r\n\x0b\x43lkResetArg\"\r\n\x0b\x43lkResetRet\"\x0f\n\rClkProgramArg\"\x0f\n\rClkProgramRet\"0\n\tClkRefArg\x12\x10\n\x08\x65xternal\x18\x01 \x01(\x08\x12\x11\n\tfrequency\x18\x02 \x01(\x01\"\x1e\n\tClkRefRet\x12\x11\n\tfrequency\x18\x01 \x01(\x01\"\"\n\rClkOutFreqArg\x12\x11\n\tfrequency\x18\x01 \x01(\x01\"\"\n\rClkOutFreqRet\x12\x11\n\tfrequency\x18\x01 \x01(\x01\"\x10\n\x0eWaitForLockArg\"&\n\x0eWaitForLockRet\x12\x14\n\x0cmilliseconds\x18\x01 \x01(\r\"\r\n\x0bIsLockedArg\"\x1d\n\x0bIsLockedRet\x12\x0e\n\x06locked\x18\x01 \x01(\x08\"7\n\x10SetSysrefModeArg\x12#\n\x04mode\x18\x01 \x01(\x0e\x32\x15.metronomo.SysrefMode\"\x12\n\x10SetSysrefModeRet\"\x0c\n\nSyncOutArg\"\x0c\n\nSyncOutRet\"\x1c\n\x0bTestFailArg\x12\r\n\x05value\x18\x01 \x01(\x05\"\r\n\x0bTestFailRet*@\n\nSysrefMode\x12\x16\n\x12SYSREF_MODE_PULSER\x10\x00\x12\x1a\n\x16SYSREF_MODE_CONTINUOUS\x10\x01\x32\xd2\x05\n\tMetronomo\x12+\n\x03Idn\x12\x11.metronomo.IdnArg\x1a\x11.metronomo.IdnRet\x12.\n\x04\x45\x63ho\x12\x12.metronomo.EchoArg\x1a\x12.metronomo.EchoRet\x12\x31\n\x05Sleep\x12\x13.metronomo.SleepArg\x1a\x13.metronomo.SleepRet\x12:\n\x08\x43lkReset\x12\x16.metronomo.ClkResetArg\x1a\x16.metronomo.ClkResetRet\x12@\n\nClkProgram\x12\x18.metronomo.ClkProgramArg\x1a\x18.metronomo.ClkProgramRet\x12\x34\n\x06\x43lkRef\x12\x14.metronomo.ClkRefArg\x1a\x14.metronomo.ClkRefRet\x12@\n\nClkOutFreq\x12\x18.metronomo.ClkOutFreqArg\x1a\x18.metronomo.ClkOutFreqRet\x12\x43\n\x0bWaitForLock\x12\x19.metronomo.WaitForLockArg\x1a\x19.metronomo.WaitForLockRet\x12:\n\x08IsLocked\x12\x16.metronomo.IsLockedArg\x1a\x16.metronomo.IsLockedRet\x12I\n\rSetSysrefMode\x12\x1b.metronomo.SetSysrefModeArg\x1a\x1b.metronomo.SetSysrefModeRet\x12\x37\n\x07SyncOut\x12\x15.metronomo.SyncOutArg\x1a\x15.metronomo.SyncOutRet\x12:\n\x08TestFail\x12\x16.metronomo.TestFailArg\x1a\x16.metronomo.TestFailRetb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'metronomo_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SYSREFMODE']._serialized_start=746
  _globals['_SYSREFMODE']._serialized_end=810
  _globals['_IDNARG']._serialized_start=30
  _globals['_IDNARG']._serialized_end=38
  _globals['_IDNRET']._serialized_start=41
  _globals['_IDNRET']._serialized_end=179
  _globals['_ECHOARG']._serialized_start=181
  _globals['_ECHOARG']._serialized_end=205
  _globals['_ECHORET']._serialized_start=207
  _globals['_ECHORET']._serialized_end=231
  _globals['_SLEEPARG']._serialized_start=233
  _globals['_SLEEPARG']._serialized_end=260
  _globals['_SLEEPRET']._serialized_start=262
  _globals['_SLEEPRET']._serialized_end=272
  _globals['_CLKRESETARG']._serialized_start=274
  _globals['_CLKRESETARG']._serialized_end=287
  _globals['_CLKRESETRET']._serialized_start=289
  _globals['_CLKRESETRET']._serialized_end=302
  _globals['_CLKPROGRAMARG']._serialized_start=304
  _globals['_CLKPROGRAMARG']._serialized_end=319
  _globals['_CLKPROGRAMRET']._serialized_start=321
  _globals['_CLKPROGRAMRET']._serialized_end=336
  _globals['_CLKREFARG']._serialized_start=338
  _globals['_CLKREFARG']._serialized_end=386
  _globals['_CLKREFRET']._serialized_start=388
  _globals['_CLKREFRET']._serialized_end=418
  _globals['_CLKOUTFREQARG']._serialized_start=420
  _globals['_CLKOUTFREQARG']._serialized_end=454
  _globals['_CLKOUTFREQRET']._serialized_start=456
  _globals['_CLKOUTFREQRET']._serialized_end=490
  _globals['_WAITFORLOCKARG']._serialized_start=492
  _globals['_WAITFORLOCKARG']._serialized_end=508
  _globals['_WAITFORLOCKRET']._serialized_start=510
  _globals['_WAITFORLOCKRET']._serialized_end=548
  _globals['_ISLOCKEDARG']._serialized_start=550
  _globals['_ISLOCKEDARG']._serialized_end=563
  _globals['_ISLOCKEDRET']._serialized_start=565
  _globals['_ISLOCKEDRET']._serialized_end=594
  _globals['_SETSYSREFMODEARG']._serialized_start=596
  _globals['_SETSYSREFMODEARG']._serialized_end=651
  _globals['_SETSYSREFMODERET']._serialized_start=653
  _globals['_SETSYSREFMODERET']._serialized_end=671
  _globals['_SYNCOUTARG']._serialized_start=673
  _globals['_SYNCOUTARG']._serialized_end=685
  _globals['_SYNCOUTRET']._serialized_start=687
  _globals['_SYNCOUTRET']._serialized_end=699
  _globals['_TESTFAILARG']._serialized_start=701
  _globals['_TESTFAILARG']._serialized_end=729
  _globals['_TESTFAILRET']._serialized_start=731
  _globals['_TESTFAILRET']._serialized_end=744
  _globals['_METRONOMO']._serialized_start=813
  _globals['_METRONOMO']._serialized_end=1535
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SysrefMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    SYSREF_MODE_PULSER: _ClassVar[SysrefMode]
    SYSREF_MODE_CONTINUOUS: _ClassVar[SysrefMode]
SYSREF_MODE_PULSER: SysrefMode
SYSREF_MODE_CONTINUOUS: SysrefMode

class IdnArg(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class IdnRet(_message.Message):
    __slots__ = ("name", "version", "serial_number", "rpi_model", "rpi_soc", "rpi_serial", "pcb_serial")
    NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    SERIAL_NUMBER_FIELD_NUMBER: _ClassVar[int]
    RPI_MODEL_FIELD_NUMBER: _ClassVar[int]
    RPI_SOC_FIELD_NUMBER: _ClassVar[int]
    RPI_SERIAL_FIELD_NUMBER: _ClassVar[int]
    PCB_SERIAL_FIELD_NUMBER: _ClassVar[int]
    name: str
    version: str
    serial_number: str
    rpi_model: str
    rpi_soc: str
    rpi_serial: str
    pcb_serial: str
    def __init__(self, name: _Optional[str] = ..., version: _Optional[str] = ..., serial_number: _Optional[str] = ..., rpi_model: _Optional[str] = ..., rpi_soc: _Optional[str] = ..., rpi_serial: _Optional[str] = ..., pcb_serial: _Optional[str] = ...) -> None: ...

class EchoArg(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: int
    def __init__(self, value: _Optional[int] = ...) -> None: ...

class EchoRet(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: int
    def __init__(self, value: _Optional[int] = ...) -> None: ...

class SleepArg(_message.Message):
    __slots__ = ("seconds",)
    SECONDS_FIELD_NUMBER: _ClassVar[int]
    seconds: float
    def __init__(self, seconds: _Optional[float] = ...) -> None: ...

class SleepRet(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ClkResetArg(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ClkResetRet(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ClkProgramArg(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ClkProgramRet(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ClkRefArg(_message.Message):
    __slots__ = ("external", "frequency")
    EXTERNAL_FIELD_NUMBER: _ClassVar[int]
    FREQUENCY_FIELD_NUMBER: _ClassVar[int]
    external: bool
    frequency: float
    def __init__(self, external: bool = ..., frequency: _Optional[float] = ...) -> None: ...

class ClkRefRet(_message.Message):
    __slots__ = ("frequency",)
    FREQUENCY_FIELD_NUMBER: _ClassVar[int]
    frequency: float
    def __init__(self, frequency: _Optional[float] = ...) -> None: ...

class ClkOutFreqArg(_message.Message):
    __slots__ = ("frequency",)
    FREQUENCY_FIELD_NUMBER: _ClassVar[int]
    frequency: float
    def __init__(self, frequency: _Optional[float] = ...) -> None: ...

class ClkOutFreqRet(_message.Message):
    __slots__ = ("frequency",)
    FREQUENCY_FIELD_NUMBER: _ClassVar[int]
    frequency: float
    def __init__(self, frequency: _Optional[float] = ...) -> None: ...

class WaitForLockArg(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class WaitForLockRet(_message.Message):
    __slots__ = ("milliseconds",)
    MILLISECONDS_FIELD_NUMBER: _ClassVar[int]
    milliseconds: int
    def __init__(self, milliseconds: _Optional[int] = ...) -> None: ...

class IsLockedArg(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class IsLockedRet(_message.Message):
    __slots__ = ("locked",)
    LOCKED_FIELD_NUMBER: _ClassVar[int]
    locked: bool
    def __init__(self, locked: bool = ...) -> None: ...

class SetSysrefModeArg(_message.Message):
    __slots__ = ("mode",)
    MODE_FIELD_NUMBER: _ClassVar[int]
    mode: SysrefMode
    def __init__(self, mode: _Optional[_Union[SysrefMode, str]] = ...) -> None: ...

class SetSysrefModeRet(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SyncOutArg(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SyncOutRet(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class TestFailArg(_message.Message):
    __slots__ = ("value",)
    VALUE_FIELD_NUMBER: _ClassVar[int]
    value: int
    def __init__(self, value: _Optional[int] = ...) -> None: ...

class TestFailRet(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...



# Copyright (c) 2023-2026 Intermodulation Products AB

"""Collection of functions that might be useful, or might not."""

import sys


def eprint(*objects):
    print(*objects, file=sys.stderr)


def _colored(aec: str, msg: str) -> str:
    return f"{aec}{msg}{_AEC_RESET}"


def eprint_error(title: str, body: str, warn: bool = False, inline: bool = False):
    aec = _AEC_YELLOW if warn else _AEC_RED
    if not title:
        eprint(_colored(aec, body))
    else:
        if inline:
            wide = f"{title}:"
            eprint(f"{_colored(aec, wide)} {body}")
        else:
            wide = f" {title} "
            eprint(_colored(aec, f"{wide:-^80}"))
            eprint(body)
            eprint(_colored(aec, f"{'':-^80}"))


_AEC_YELLOW = "\x1b[1;33m"
_AEC_RED = "\x1b[1;31m"
_AEC_RESET = "\x1b[0m"
